import logging
from abc import ABC, abstractmethod
from typing import Callable, Iterable

import dill

from mhagenta.utils.common import Message, MHABase, StatusReport, AgentCmd, ModuleTypes, AgentTime, LoggerExtras, \
    DEFAULT_LOG_FORMAT
from mhagenta.utils.common.typing import MsgProcessorCallback


class Connector(MHABase, ABC):
    def __init__(self,
                 agent_id: str,
                 sender_id: str,
                 agent_time: AgentTime,
                 log_tags: list[str] = '',
                 log_level: int | str = logging.DEBUG,
                 log_format: str = DEFAULT_LOG_FORMAT,
                 *args, **kwargs
                 ) -> None:
        super().__init__(
            agent_id=agent_id,
            log_tags=log_tags,
            log_level=log_level,
            log_format=log_format
        )
        self._id = self.__class__.__name__
        self._sender_id = sender_id
        self._time = agent_time

    @abstractmethod
    async def start(self) -> None:
        pass

    @abstractmethod
    async def stop(self) -> None:
        pass

    @abstractmethod
    async def subscribe_to_in_channel(self, sender: str, channel: str, callback: MsgProcessorCallback, **kwargs) -> None:
        pass

    @abstractmethod
    async def register_out_channel(self, recipient: str, channel: str, **kwargs) -> None:
        pass

    @abstractmethod
    async def subscribe_to_cmds(self, callback: Callable[[AgentCmd], None], **kwargs) -> None:
        pass

    @abstractmethod
    async def register_cmd_out_channel(self, **kwargs) -> None:
        pass

    @abstractmethod
    async def subscribe_to_statuses(self, callback: Callable[[StatusReport], None], **kwargs) -> None:
        pass

    @abstractmethod
    async def register_status_out_channel(self, **kwargs) -> None:
        pass

    @abstractmethod
    def send(self, recipient: str, channel: str, msg: Message, **kwargs) -> None:
        pass

    @abstractmethod
    def cmd(self,
            cmd: AgentCmd,
            module_types: str | Iterable[str] = ModuleTypes.ALL,
            module_ids: str | Iterable[str] = ModuleTypes.ALL,
            **kwargs) -> None:
        pass

    @abstractmethod
    def status(self, status: StatusReport, **kwargs) -> None:
        pass

    @staticmethod
    def encode_msg(msg: Message) -> bytes:
        return dill.dumps(msg)

    @staticmethod
    def decode_msg(msg: bytes) -> Message:
        return dill.loads(msg)

    @property
    def _logger_extras(self) -> LoggerExtras | None:
        return LoggerExtras(
            agent_time=self._time.agent,
            mod_time=self._time.module,
            exec_time=str(self._time.exec) if self._time.exec is not None else '-',
            tags=self.log_tag_str
        )




import logging
import sys
import time
from abc import ABC, abstractmethod
from typing import Any, ClassVar, Callable, Self, Iterable, Optional
from uuid import uuid4

from pydantic.dataclasses import dataclass

from mhagenta.utils.common.logging import ILogging
from mhagenta.utils.common.logging import DEFAULT_FORMAT as DEFAULT_LOG_FORMAT


class MHABase(ILogging, ABC):
    def __init__(self,
                 agent_id: str,
                 log_id: str | None = None,
                 log_tags: list[str] | None = None,
                 log_level: int | str = logging.DEBUG,
                 log_format: str = DEFAULT_LOG_FORMAT
                 ) -> None:
        self._agent_id = agent_id
        self._log_id = self.__class__.__name__ if log_id is None else log_id
        self._log_tags = log_tags[:] if log_tags else [agent_id]
        self._log_tags.append(self._log_id)
        self._log_level = log_level
        self._log_format = log_format

        self._local_logger = logging.getLogger('.'.join(self._log_tags))
        self._local_logger.setLevel(log_level)
        if not self._local_logger.hasHandlers():
            handler = logging.StreamHandler(sys.stdout)
            handler.setFormatter(logging.Formatter(log_format))
            self._local_logger.addHandler(handler)

    @abstractmethod
    async def initialize(self) -> None:
        pass

    @property
    def log_id(self) -> str:
        return self._log_id

    @property
    def log_tag_str(self) -> str:
        return f'[{"][".join(self._log_tags)}]'

    @property
    def _logger(self) -> logging.Logger:
        return self._local_logger


class AgentTime:
    """Utility class for getting timestamps for agent components.

    """
    def __init__(self, agent_start_ts: float, exec_start_ts: float | None = None, decimals: int = 4) -> None:
        self._agent_start_ts = agent_start_ts
        self._module_start_ts: float = time.time()
        self._exec_start_ts: float | None = exec_start_ts
        self._decimals = decimals

    @property
    def system(self) -> float:
        """System time in seconds.

        Returns:
            float: System time (in seconds) as provided by `time` module.

        """
        return round(time.time(), self._decimals)

    @property
    def agent(self) -> float:
        """Agent time in seconds.

        Returns:
            float: Seconds since the agent was created (i.e. since the initialization of agent's root controller).

        """
        return round(time.time() - self._agent_start_ts, self._decimals)

    @property
    def module(self) -> float:
        """Module time in seconds.

        Returns:
            float: Seconds since the initializations of the current module.
        """
        return round(time.time() - self._module_start_ts, self._decimals)

    @property
    def exec(self) -> float | None:
        """Execution time in seconds.

        Returns:
            float | None: Seconds since the synchronous start of agent execution. Is None if the module doesn't have
            the information yet, or negative if the execution start is scheduled in the future.
         """
        return round(time.time() - self._exec_start_ts, self._decimals) if self._exec_start_ts is not None else None

    @property
    def agent_start_ts(self) -> float:
        """Unix timestamp in seconds of when the synchronous agents execution start.

        Returns:
            float: Unix timestamp in seconds.

        """
        return self._agent_start_ts

    @property
    def exec_start_ts(self) -> float | None:
        """Unix timestamp in seconds of when this agent starts execution.

        Returns:
            float | None: Unix timestamp in seconds. None if not yet defined.

        """
        return self._exec_start_ts

    def set_exec_start_ts(self, exec_start_ts: float) -> None:
        """
        Pass the information on when this agent execution is scheduled to start.

        Args:
            exec_start_ts: system-level Unix timestamp (in seconds) of agent's scheduled start time.
        """
        self._exec_start_ts = exec_start_ts

    def get_exec_time(self) -> float | None:
        """
        Functional version of `exec` property. Used by execution queues.

        Returns:
            float if defined, None otherwise
        """
        return self.exec

    @staticmethod
    def sleep(secs: float) -> None:
        time.sleep(secs)

    @property
    def tuple(self) -> tuple[float, float, float, float | None]:
        """Full time information as a tuple.

        Returns:
            tuple[float, float, float, float | None]: tuple of all four types of timestamps in order: system, agent,
                module, execution

        """
        return self.system, self.agent, self.module, self.exec


class Directory:
    """Directory of all module names and types for easier communication definition.

    Created by the agent's root controller.

    """
    def __init__(self,
                 perception: Iterable[str] | str,
                 actuation: Iterable[str] | str,
                 ll_reasoning: Iterable[str] | str,
                 learning: Iterable[str] | str | None = None,
                 knowledge: Iterable[str] | str | None = None,
                 hl_reasoning: Iterable[str] | str | None = None,
                 goals: Iterable[str] | str | None = None,
                 memory: Iterable[str] | str | None = None
                 ) -> None:
        self._perception = self._process_modules(perception)
        self._actuation = self._process_modules(actuation)
        self._ll_reasoning = self._process_modules(ll_reasoning)
        self._learning = self._process_modules(learning)
        self._knowledge = self._process_modules(knowledge)
        self._hl_reasoning = self._process_modules(hl_reasoning)
        self._goals = self._process_modules(goals)
        self._memory = self._process_modules(memory)

    @property
    def perception(self) -> list[str]:
        """(Property) List of Perceptor IDs.

        Returns:
            list[str]: List of `module_id`s of all the agent `Perceptor`s

        """
        return self._perception

    @property
    def actuation(self) -> list[str]:
        """(Property) List of Actuator IDs.

        Returns:
            list[str]: List of `module_id`s of all the agent `Actuator`s

        """
        return self._actuation

    @property
    def ll_reasoning(self) -> list[str]:
        """(Property) List of Low-level reasoner IDs.

        Returns:
            list[str]: List of `module_id`s of all the agent `LLReasoner`s

        """
        return self._ll_reasoning

    @property
    def learning(self) -> list[str]:
        """(Property) List of Learner IDs.

        Returns:
            list[str]: List of `module_id`s of all the agent `Learner`s

        """
        return self._learning

    @property
    def knowledge(self) -> list[str]:
        """(Property) List of Knowledge model IDs.

        Returns:
            list[str]: List of `module_id`s of all the agent `Knowledge`s

        """
        return self._knowledge

    @property
    def hl_reasoning(self) -> list[str]:
        """(Property) List of High-level reasoner IDs.

        Returns:
            list[str]: List of `module_id`s of all the agent `HLReasoner`s

        """
        return self._hl_reasoning

    @property
    def goals(self) -> list[str]:
        """(Property) List of Goal graph IDs.

        Returns:
            list[str]: List of `module_id`s of all the agent `goal_graph`s

        """
        return self._goals

    @property
    def memory(self) -> list[str]:
        """(Property) List of Memory structure IDs.

        Returns:
            list[str]: List of `module_id`s of all the agent `Memory`s

        """
        return self._memory

    @staticmethod
    def _process_modules(module_ids: Iterable[str] | str | None) -> list[str]:
        if isinstance(module_ids, Iterable):
            modules = list(module_ids)
        elif isinstance(module_ids, str):
            modules = [module_ids]
        else:
            modules = []
        return modules


@dataclass
class Belief:
    """Pydantic dataclass for agent's beliefs.

    Attributes:
        predicate (str): predicate name.
        arguments (Any | tuple[Any, ...]): predicate's arguments.
        extras (dict[str, Any], optional): keyword dictionary of additional relevant information.

    """
    predicate: str
    arguments: Any | tuple[Any, ...]
    extras: Optional[dict[str, Any]] = None


@dataclass
class Goal:
    """Pydantic dataclass for agent's goals.

    Attributes:
        state (list[Belief]): Belief-based description of the goal state.
        misc (dict[str, Any]): keyword dictionary of additional relevant information.

    """
    state: list[Belief]
    extras: Optional[dict[str, Any]] = None

    def __init__(self, state: list[Belief], **kwargs) -> None:
        super().__init__(state=state, misc=kwargs)


@dataclass
class Observation:
    """Pydantic dataclass for agent's observations.

    Attributes:
        observation_type (str): type of the observation.
        value (Any): observed object.

    """
    value: Any
    observation_type: Optional[str] = 'Any'

    def __str__(self) -> str:
        return f'[{self.observation_type}] {self.value}'


@dataclass
class ActionStatus:
    """Pydantic dataclass for action statuses.

    Attributes:
        status (Any): execution status of the action.

    """
    status: Any


class ConnType:
    send = 'send'
    request = 'request'


@dataclass
class AgentCmd:
    START: ClassVar[str] = 'start'
    STOP: ClassVar[str] = 'stop'

    agent_id: str
    cmd: str
    args: dict[str, Any] | None = None

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if self.args is None:
            self.args = dict()


@dataclass
class StatusReport:
    CREATED: ClassVar[str] = 'CREATED'
    READY: ClassVar[str] = 'READY'
    RUNNING: ClassVar[str] = 'RUNNING'
    FINISHED: ClassVar[str] = 'FINISHED'
    ERROR: ClassVar[str] = 'ERROR'
    TIMEOUT: ClassVar[str] = 'TIMEOUT'

    agent_id: str
    module_id: str
    status: str
    ts: float
    args: dict[str, Any] | None = None

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if self.args is None:
            self.args = dict()

    def __str__(self) -> str:
        return f'{self.__class__.__name__}[{self.agent_id}.{self.module_id}]({self.status}, {self.ts}{f": {self.args}" if self.args else ""})'


@dataclass
class Message:

    body: Any | dict[str, Any]
    sender_id: str
    recipient_id: str
    ts: float | str
    performative: str
    uuid: bytes = b''
    short_uuid_format: Optional[bool] = True

    def __init__(self,
                 body: Any | dict[str, Any],
                 sender_id: str,
                 recipient_id: str,
                 ts: float | str,
                 performative: str
                 ) -> None:
        super().__init__(
            sender_id=sender_id,
            recipient_id=recipient_id,
            ts=ts,
            performative=performative,
            body=body)
        self.uuid = uuid4().bytes

    def __str__(self) -> str:
        return f'[{self.ts}][{self.sender_id}-->{self.recipient_id}]' \
               f'{f"[{self.performative}]" if self.performative else ""} ' \
               f'{self.body}'

    @property
    def id(self) -> str:
        return self.short_id if self.short_uuid_format else self.full_id

    @property
    def full_id(self) -> str:
        return self.uuid.hex()

    @property
    def short_id(self) -> str:
        return self.uuid[-6:].hex()


class Outbox(ABC):
    def __init__(self) -> None:
        self._msgs: dict[tuple[str, str, str | None], list[Any | dict[str, Any]]] = dict()

        self._recipients = list()
        self._next_recipient = -1
        self._next_content = -1

    def _add(self, recipient_id: str, performative: str, content: Any | dict[str, Any], extension: str = '') -> None:
        if (recipient_id, performative) in self._msgs:
            self._msgs[recipient_id, performative, extension].append(content)
        else:
            self._msgs[recipient_id, performative, extension] = [content]

    def __iter__(self) -> Self:
        self._recipients = list(self._msgs.keys())
        self._next_recipient = 0
        self._next_content = 0
        return self

    def __next__(self) -> tuple[str, str, str, Any | dict[str, Any]]:
        if self._next_content >= len(self._msgs[self._recipients[self._next_recipient]]):
            self._next_recipient += 1
            self._next_content = 0

        if self._next_recipient >= len(self._msgs):
            raise StopIteration

        recipient_id, performative, extension = self._recipients[self._next_recipient]
        content = self._msgs[self._recipients[self._next_recipient]][self._next_content]
        self._next_content += 1
        return recipient_id, performative, extension, content

    def __bool__(self) -> bool:
        return bool(self._msgs)

    def __str__(self) -> str:
        messages = list()
        for recipient, content_list in self._msgs.items():
            for content in content_list:
                messages.append(f'[TO: {recipient[0]}][{recipient[1]}{f"/{recipient[2]}" if recipient[2] else ""}]({content})')
        return f'Outbox({", ".join(messages)})'

    def __repr__(self) -> str:
        return self.__str__()

    def clear(self) -> None:
        self._msgs.clear()
        self._recipients = list()
        self._next_recipient = -1
        self._next_content = -1


class State[T: Outbox]:
    """Container for module's internal state enriched with additional information and functionality.

    State creation is handled by the agent's root controller. Can be used to access agent time, directory of agent
    module IDs, and to send out messages to other modules.

    T resolves to a module-specific outbox class for convenient hinting when sending messages to other modules.

    """
    def __init__(self, agent_id: str, module_id: str, time_func: Callable[[], float], directory: Directory, outbox: T, **kwargs) -> None:
        self._agent_id = agent_id
        self._module_id = module_id
        self._time_func = time_func
        self._directory = directory

        self._outbox = outbox

        self._custom_fields = set(kwargs.keys())
        self.__dict__.update(kwargs)

    @property
    def agent_id(self) -> str:
        """Unique ID of the parent agent.

        Returns:
            str: agent_id specified at the agent's creation.

        """
        return self._agent_id

    @property
    def module_id(self) -> str:
        """Unique (within the scope of the parent agent) identifier of the current module.

        Returns:
            str: module_id specified in the module base object.

        """
        return self._module_id

    @property
    def time(self) -> float:
        """Agent's execution current time.

        Returns:
            float: time in seconds since the start of the agent's execution.

        """
        return self._time_func()

    @property
    def directory(self) -> Directory:
        """Agent's directory of module IDs.

        Returns:
            Directory: directory object containing information about agent's module types and IDs.

        """
        return self._directory

    @property
    def outbox(self) -> T:
        """Outbox object for this module. Use its methods to send messages to other modules.

        Returns:
            T: module-specific outbox object.

        """
        return self._outbox

    def _clear_outbox(self) -> None:
        self._outbox.clear()

    def dump(self) -> dict[str, Any]:
        return {field: self.__dict__[field] for field in self._custom_fields}

    def load(self, **kwargs) -> None:
        self._custom_fields.update(kwargs.keys())
        self.__dict__.update(kwargs)

    def __str__(self) -> str:
        return str(self.dump())

    def __repr__(self) -> str:
        return str(self.dump())


class ModuleTypes:
    AGENT = 'Agent'
    ALL = 'All'
    PERCEPTOR = 'Perceptor'
    ACTUATOR = 'Actuator'
    LLREASONER = 'LLReasoner'
    LEARNER = 'Learner'
    KNOWLEDGE = 'Knowledge'
    HLREASONER = 'HLReasoner'
    GOALGRAPH = 'GoalGraph'
    MEMORY = 'Memory'



import asyncio
import logging
import shutil
import time
from asyncio import TaskGroup
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterable, Literal

import dill
import docker
from docker.errors import NotFound
from docker.models.containers import Container
from docker.models.images import Image

import mhagenta
from mhagenta.bases import *
from mhagenta.containers import *
from mhagenta.core.connection import Connector, RabbitMQConnector
from mhagenta.utils.common import DEFAULT_LOG_FORMAT


@dataclass
class AgentEntry:
    agent_id: str
    kwargs: dict[str, Any]
    dir: Path | None = None
    save_dir: Path | None = None
    image: Image | None = None
    container: Container | None = None
    port_mapping: dict[int, int] | None = None
    num_copies: int = 1


class Orchestrator:
    """Orchestrator class that handles MHAgentA execution.

    Orchestrator handles definition of agents and their consequent containerization and deployment. It also allows you
    to define default parameters shared by all the agents handles by it (can be overridden by individual agents)

    """
    SAVE_SUBDIR = 'out/save'
    LOG_CHECK_FREQ = 1.

    def __init__(self,
                 save_dir: str | Path,
                 port_mapping: dict[int, int] | None = None,
                 step_frequency: float = 1.,
                 status_frequency: float = 5.,
                 control_frequency: float = -1.,
                 exec_start_time: float | None = None,
                 agent_start_delay: float = 5.,
                 exec_duration: float = 60.,
                 save_format: Literal['json', 'dill'] = 'json',
                 resume: bool = False,
                 log_level: int = logging.INFO,
                 log_format: str | None = None,
                 status_msg_format: str = '[status_upd]::{}',

                 module_start_delay: float = 2.,
                 connector_cls: type[Connector] = RabbitMQConnector,
                 connector_kwargs: dict[str, Any] | None = None,
                 prerelease: bool = False
                 ) -> None:
        """
        Constructor method for Orchestrator.

        Args:
            save_dir (str | Path): Root directory for storing agents' states, logs, and temporary files.
            port_mapping (dict[int, int], optional): Mapping between internal docker container ports and host ports.
            step_frequency (float, optional, default=1.0): For agent modules with periodic step functions, the
                frequency in seconds of the step function calls that modules will try to maintain (unless their
                execution takes longer, then the next iteration will be scheduled without a time delay).
            status_frequency (float, optional, default=10.0): Frequency with which agent modules will report their
                statuses to the agent's root controller (error statuses will be reported immediately, regardless of
                the value).
            control_frequency (float, optional): Frequency of agent modules' internal clock when there's no tasks
                pending. If undefined or not positive, there will be no scheduling delay.
            exec_start_time (float, optional): Unix timestamp in seconds of when the agent's execution will try to
                start (unless agent's initialization takes longer than that; in this case the agent will start
                execution as soon as it finishes initializing). If not specified, agents will start execution
                immediately after their initialization.
            agent_start_delay (float, optional, default=5.0): Delay in seconds before agents starts execution. Use when
                `exec_start_time` is not defined to stage synchronous agents start at `agent_start_delay` seconds from
                the `run()` or `arun()` call.
            exec_duration (float, optional, default=60.0):  Time limit for agent execution in seconds. All agents will
                timeout after this time.
            save_format (Literal['json', 'dill'], optional, default='json'): Format of agent modules state save files. JSON
                is more restrictive of what fields the states can include, but it is readable by humans.
            resume (bool, optional, default=False): Specifies whether to use save module states when restarting an
                agent with preexisting ID.
            log_level (int, optional, default=logging.INFO): Logging level.
            log_format (str, optional): Format of agent log messages. Defaults to
                `[%(agent_time)f|%(mod_time)f|%(exec_time)s][%(levelname)s]::%(tags)s::%(message)s`
            status_msg_format (str, optional): Format of agent status messages for external monitoring. Defaults to
                `[status_upd]::{}`
            connector_cls (type[Connector], optional, default=RabbitMQConnector): internal connector class that
                implements communication between modules. MHAgentA agents use RabbitMQ-based connectors by default.
            connector_kwargs (dict[str, Any], optional): Additional keyword arguments for connector. For
                RabbitMQConnector, the default parameters are: {`host`: 'localhost', `port`: 5672, `prefetch_count`: 1}.
            prerelease (bool, optional, default=False): Specifies whether to allow agents to use the latest prerelease
                version of mhagenta while building the container.
        """
        self._agents: dict[str, AgentEntry] = dict()

        if isinstance(save_dir, str):
            save_dir = Path(save_dir)
        if not save_dir.exists():
            save_dir.mkdir(parents=True, exist_ok=True)
        self._save_dir = save_dir

        self._package_dir = str(Path(mhagenta.__file__).parent.absolute())

        self._connector_cls = connector_cls if connector_cls else RabbitMQConnector
        if connector_kwargs is None and connector_cls == RabbitMQConnector:
            self._connector_kwargs = {
                'host': 'localhost',
                'port': 5672,
                'prefetch_count': 1
            }
        else:
            self._connector_kwargs = connector_kwargs

        self._port_mapping = port_mapping if port_mapping else {}

        self._step_frequency = step_frequency
        self._status_frequency = status_frequency
        self._control_frequency = control_frequency
        self._module_start_delay = module_start_delay
        self._exec_start_time = exec_start_time
        self._exec_duration_sec = exec_duration
        self._agent_start_delay = agent_start_delay

        self._save_format = save_format
        self._resume = resume

        self._log_level = log_level
        self._log_format = log_format if log_format else DEFAULT_LOG_FORMAT
        self._status_msg_format = status_msg_format

        self._prerelease = prerelease

        self._start_time: float = -1.
        self._simulation_end_ts = -1.

        self._docker_client: docker.DockerClient | None = None
        self._rabbitmq_image: Image | None = None
        self._base_image: Image | None = None

        self._task_group: TaskGroup | None = None
        self._force_run = False

        self._docker_init()

        self._running = False
        self._stopping = False
        self._all_stopped = False

    def _docker_init(self) -> None:
        self._docker_client = docker.from_env()

    def add_agent(self,
                  agent_id: str,
                  perceptors: Iterable[PerceptorBase] | PerceptorBase,
                  actuators: Iterable[ActuatorBase] | ActuatorBase,
                  ll_reasoners: Iterable[LLReasonerBase] | LLReasonerBase,
                  learners: Iterable[LearnerBase] | LearnerBase | None = None,
                  knowledge: Iterable[KnowledgeBase] | KnowledgeBase | None = None,
                  hl_reasoners: Iterable[HLReasonerBase] | HLReasonerBase | None = None,
                  goal_graphs: Iterable[GoalGraphBase] | GoalGraphBase | None = None,
                  memory: Iterable[MemoryBase] | MemoryBase | None = None,
                  num_copies: int = 1,
                  step_frequency: float | None = None,
                  status_frequency: float | None = None,
                  control_frequency: float | None = None,
                  exec_start_time: float | None = None,
                  start_delay: float = 0.,
                  exec_duration: float | None = None,
                  resume: bool | None = None,
                  log_level: int | None = None,
                  port_mapping: dict[int, int] | None = None,
                  connector_cls: type[Connector] | None = None,
                  connector_kwargs: dict[str, Any] | None = None
                  ) -> None:
        """Define an agent model to be added to the execution.

        This can be either a single agent, a set of identical agents following the same structure model.

        Args:
            agent_id (str): A unique identifier for the agent.
            perceptors (Iterable[PerceptorBase] | PerceptorBase): Definition(s) of agent's perceptor(s).
            actuators (Iterable[ActuatorBase] | ActuatorBase): Definition(s) of agent's actuator(s).
            ll_reasoners (Iterable[LLReasonerBase] | LLReasonerBase): Definition(s) of agent's ll_reasoner(s).
            learners (Iterable[LearnerBase] | LearnerBase, optional): Definition(s) of agent's learner(s).
            knowledge (Iterable[KnowledgeBase] | KnowledgeBase, optional): Definition(s) of agent's knowledge model(s).
            hl_reasoners (Iterable[HLReasonerBase] | HLReasonerBase, optional): Definition(s) of agent's hl_reasoner(s).
            goal_graphs (Iterable[GoalGraphBase] | GoalGraphBase, optional): Definition(s) of agent's goal_graph(s).
            memory (Iterable[MemoryBase] | MemoryBase, optional): Definition(s) of agent's memory structure(s).
            num_copies (int, optional, default=1): Number of copies of the agent to instantiate at runtime.
            step_frequency (float, optional): For agent modules with periodic step functions, the frequency in seconds
                of the step function calls that modules will try to maintain (unless their execution takes longer, then
                the next iteration will be scheduled without a time delay). Defaults to the Orchestrator's
                `step_frequency`.
            status_frequency (float, optional): Frequency with which agent modules will report their statuses to the
                agent's root controller (error statuses will be reported immediately, regardless of the value).
                Defaults to the Orchestrator's `status_frequency`.
            control_frequency (float, optional): Frequency of agent modules' internal clock when there's no tasks
                pending. If undefined or not positive, there will be no scheduling delay. Defaults to the
                Orchestrator's `control_frequency`.
            exec_start_time (float, optional): Unix timestamp in seconds of when the agent's execution will try to
                start (unless agent's initialization takes longer than that; in this case the agent will start
                execution as soon as it finishes initializing). Defaults to the Orchestrator's `exec_start_time`.
            start_delay (float, optional, default=0.0): A time offset from the global execution time start when this agent will
                attempt to start its own execution.
            exec_duration (float, optional): Time limit for agent execution in seconds. The agent will timeout after
                this time. Defaults to the Orchestrator's `exec_duration`.
            resume (bool, optional): Specifies whether to use save module states when restarting an agent with
                preexisting ID. Defaults to the Orchestrator's `resume`.
            log_level (int, optional):  Logging level for the agent. Defaults to the Orchestrator's `log_level`.
            port_mapping (dict[int, int], optional): Mapping between internal docker container ports and host ports.
                Defaults to the Orchestrator's `port_mapping`.
            connector_cls (type[Connector], optional): internal connector class that implements communication between
                modules. Defaults to the Orchestrator's `connector_cls`.
            connector_kwargs (dict[str, Any], optional): Additional keyword arguments for connector. Defaults to
                the Orchestrator's `connector_kwargs`.

        """
        kwargs = {
            'agent_id': agent_id,
            'connector_cls': connector_cls if connector_cls else self._connector_cls,
            'perceptors': perceptors,
            'actuators': actuators,
            'll_reasoners': ll_reasoners,
            'learners': learners,
            'knowledge': knowledge,
            'hl_reasoners': hl_reasoners,
            'goal_graphs': goal_graphs,
            'memory': memory,
            'connector_kwargs': connector_kwargs if connector_kwargs else self._connector_kwargs,
            'step_frequency': self._step_frequency if step_frequency is None else step_frequency,
            'status_frequency': self._status_frequency if status_frequency is None else status_frequency,
            'control_frequency': self._control_frequency if control_frequency is None else control_frequency,
            'exec_start_time': self._exec_start_time if exec_start_time is None else exec_start_time,
            'start_delay': start_delay,
            'exec_duration': self._exec_duration_sec if exec_duration is None else exec_duration,
            'save_dir': f'/{self.SAVE_SUBDIR}/{agent_id}',
            'save_format': self._save_format,
            'resume': self._resume if resume is None else resume,
            'log_level': self._log_level if log_level is None else log_level,
            'log_format': self._log_format,
            'status_msg_format': self._status_msg_format
        }

        self._agents[agent_id] = AgentEntry(
            agent_id=agent_id,
            port_mapping=port_mapping if port_mapping else self._port_mapping,
            num_copies=num_copies,
            kwargs=kwargs
        )
        if self._task_group is not None:
            self._task_group.create_task(self._run_agent(self._agents[agent_id], force_run=self._force_run))

    def _docker_build_base(self,
                           rabbitmq_image_name: str = 'mha-rabbitmq',
                           mha_base_image_name: str = 'mha-base',
                           version_tag: str | None = None
                           ) -> None:
        if version_tag is None:
            version_tag = CONTAINER_VERSION
        print(f'===== BUILDING RABBITMQ BASE IMAGE: {rabbitmq_image_name}:{version_tag} =====')
        if self._rabbitmq_image is None:
            self._rabbitmq_image, _ = (
                self._docker_client.images.build(path=RABBIT_IMG_PATH,
                                                 tag=f'{rabbitmq_image_name}:{version_tag}',
                                                 rm=True,
                                                 quiet=False
                                                 ))

        print(f'===== BUILDING AGENT BASE IMAGE: {mha_base_image_name}:{version_tag} =====')
        if self._base_image is None:
            try:
                self._base_image = self._docker_client.images.list(name=f'{mha_base_image_name}:{version_tag}')[0]
            except IndexError:
                build_dir = self._save_dir.absolute() / 'tmp/'
                try:
                    shutil.copytree(BASE_IMG_PATH, build_dir)
                    shutil.copytree(self._package_dir, build_dir / 'mhagenta')

                    self._base_image, _ = (
                        self._docker_client.images.build(path=str(build_dir),
                                                         buildargs={
                                                             'SRC_IMAGE': rabbitmq_image_name,
                                                             'SRC_VERSION': version_tag,
                                                             'PRE_VERSION': "true" if self._prerelease else "false"
                                                         },
                                                         tag=f'{mha_base_image_name}:{version_tag}',
                                                         rm=True,
                                                         quiet=False
                                                         ))
                except Exception as ex:
                    shutil.rmtree(build_dir, ignore_errors=True)
                    raise ex
                shutil.rmtree(build_dir)

    def _docker_build_agent(self,
                            agent: AgentEntry
                            ) -> None:
        print(f'===== BUILDING AGENT IMAGE: mhagent:{agent.agent_id} =====')
        agent_dir = self._save_dir.absolute() / agent.agent_id
        if self._force_run and agent_dir.exists():
            shutil.rmtree(agent_dir)

        (agent_dir / 'out/').mkdir(parents=True)
        agent.dir = agent_dir
        agent.save_dir = agent_dir / 'out' / 'save' / agent.agent_id

        build_dir = agent_dir / 'tmp/'
        shutil.copytree(AGENT_IMG_PATH, build_dir.absolute())
        (build_dir / 'src').mkdir(parents=True, exist_ok=True)
        shutil.copy(Path(mhagenta.core.__file__).parent.absolute() / 'agent_launcher.py', (build_dir / 'src' / 'agent_launcher.py').absolute())
        shutil.copy(Path(mhagenta.__file__).parent.absolute() / 'scripts' / 'start.sh', (build_dir / 'src' / 'start.sh').absolute())

        if agent.kwargs['exec_start_time'] is None:
            agent.kwargs['exec_start_time'] = self._start_time
        agent.kwargs['exec_start_time'] += self._agent_start_delay

        end_estimate = agent.kwargs['exec_start_time'] + agent.kwargs['start_delay'] + agent.kwargs['exec_duration']
        if self._simulation_end_ts < end_estimate:
            self._simulation_end_ts = end_estimate

        with open((build_dir / 'src' / 'agent_params').absolute(), 'wb') as f:
            dill.dump(agent.kwargs, f, recurse=True)

        base_tag = self._base_image.tags[0].split(':')
        agent.image, _ = self._docker_client.images.build(path=str(build_dir.absolute()),
                                                          buildargs={
                                                              'SRC_IMAGE': base_tag[0],
                                                              'SRC_VERSION': base_tag[1]
                                                          },
                                                          tag=f'mhagent:{agent.agent_id}',
                                                          rm=True,
                                                          quiet=False
                                                          )
        shutil.rmtree(build_dir)

    async def _run_agent(self,
                         agent: AgentEntry,
                         force_run: bool = False
                         ) -> None:
        if agent.num_copies == 1:
            print(f'===== RUNNING AGENT IMAGE \"mhagent:{agent.agent_id}\" AS CONTAINER \"{agent.agent_id}\" =====')
        else:
            print(f'===== RUNNING AGENT IMAGE \"mhagent:{agent.agent_id}\" AS '
                  f'{agent.num_copies} CONTAINERS \"{agent.agent_id}_#\" =====')
        for i in range(agent.num_copies):
            if agent.num_copies == 1:
                agent_name = agent.agent_id
                agent_dir = (agent.dir / "out").absolute()
            else:
                agent_name = f'{agent.agent_id}_{i}'
                agent_dir = (agent.dir / str(i) / "out").absolute()

            agent_dir.mkdir(parents=True, exist_ok=True)
            try:
                container = self._docker_client.containers.get(agent_name)
                if force_run:
                    container.remove(force=True)
                else:
                    raise NameError(f'Container {agent_name} already exists')
            except NotFound:
                pass

            agent.container = self._docker_client.containers.run(image=agent.image,
                                                                 detach=True,
                                                                 name=agent_name,
                                                                 environment={"AGENT_ID": agent_name},
                                                                 volumes={
                                                                     str(agent_dir): {'bind': '/out', 'mode': 'rw'}
                                                                 },
                                                                 extra_hosts={'host.docker.internal': 'host-gateway'},
                                                                 ports=agent.port_mapping)

    async def arun(self,
                   rabbitmq_image_name: str = 'mha-rabbitmq',
                   hagent_base_image_name: str = 'mha-base',
                   force_run: bool = False
                   ) -> None:
        """Run all the agents as an async method. Use in case you want to control the async task loop yourself.

        Args:
            rabbitmq_image_name (str, optional, default='mha-rabbitmq'): The name of the base MHAgentA RabbitMQ image.
            hagent_base_image_name (str, optional, default='mha-base'): The name of the base MHAgentA agent image.:
            force_run (bool, optional, default=False): In case containers with some of the specified agent IDs exist,
                specify whether to force remove the old container to run the new ones. Otherwise, an exception will be
                raised.

        Raises:
            NameError: Raised if a container for one of the specified agent IDs already exists and `force_run` is False.

        """
        if self._base_image is None:
            self._docker_build_base(rabbitmq_image_name=rabbitmq_image_name,
                                    mha_base_image_name=hagent_base_image_name)

        self._force_run = force_run
        for agent in self._agents.values():
            self._docker_build_agent(agent)

        self._running = True
        self._start_time = time.time()
        async with asyncio.TaskGroup() as tg:
            self._task_group = tg
            for agent in self._agents.values():
                tg.create_task(self._run_agent(agent, force_run=force_run))
                tg.create_task(self._simulation_end_timer())
                tg.create_task(self._read_logs(agent))
        self._running = False
        for agent in self._agents.values():
            agent.container.remove()
        print('===== EXECUTION FINISHED =====')

    def run(self,
            rabbitmq_image_name: str = 'mha-rabbitmq',
            hagent_base_image_name: str = 'mha-base',
            force_run: bool = False
            ) -> None:
        """Run all the agents.

        Args:
            rabbitmq_image_name (str, optional, default='mha-rabbitmq'): The name of the base MHAgentA RabbitMQ image.
            hagent_base_image_name (str, optional, default='mha-base'): The name of the base MHAgentA agent image.:
            force_run (bool, optional, default=False): In case containers with some of the specified agent IDs exist,
                specify whether to force remove the old container to run the new ones. Otherwise, an exception will be
                raised.

        Raises:
            NameError: Raised if a container for one of the specified agent IDs already exists and `force_run` is False.

        """
        asyncio.run(self.arun(
            rabbitmq_image_name=rabbitmq_image_name,
            hagent_base_image_name=hagent_base_image_name,
            force_run=force_run
        ))

    @staticmethod
    def _agent_stopped(agent: AgentEntry) -> bool:
        agent.container.reload()
        return agent.container.status == 'exited'

    @property
    def _agents_stopped(self) -> bool:
        if self._all_stopped:
            return True
        for agent in self._agents.values():
            if not self._agent_stopped(agent):
                return False
        self._all_stopped = True
        return True

    async def _simulation_end_timer(self) -> None:
        await asyncio.sleep(self._simulation_end_ts - time.time())
        self._stopping = True

    def _add_log(self, log: str | bytes) -> None:
        if isinstance(log, bytes):
            log = log.decode()
        print(log.strip('\n\r'))

    async def _read_logs(self, agent: AgentEntry) -> None:
        logs = self._docker_client.containers.get(agent.container.id).logs(stdout=True, stderr=True, stream=True, follow=True)

        while True:
            if self._stopping and self._agents_stopped:
                break
            for line in logs:
                self._add_log(line)
            await asyncio.sleep(self.LOG_CHECK_FREQ)

    def __getitem__(self, agent_id: str) -> AgentEntry:
        return self._agents[agent_id]



import asyncio
import logging
from typing import Iterable, Callable

from mhagenta.utils.common import MHABase, AgentTime, ModuleTypes, DEFAULT_LOG_FORMAT
from mhagenta.core.connection.connector import Connector
from mhagenta.utils import AgentCmd, StatusReport, LoggerExtras


class RootMessenger(MHABase):
    def __init__(self,
                 connector_cls: type[Connector],
                 agent_id: str,
                 agent_time: AgentTime,
                 status_callback: Callable[[StatusReport], None],
                 log_tags: list[str],
                 log_level: int | str = logging.DEBUG,
                 log_format: str = DEFAULT_LOG_FORMAT,
                 **kwargs
                 ) -> None:
        super().__init__(
            agent_id=agent_id,
            log_tags=log_tags,
            log_level=log_level,
            log_format=log_format
        )

        self._time = agent_time
        self._status_callback = status_callback

        self._connector = connector_cls(
            agent_id=agent_id,
            agent_time=agent_time,
            sender_id=f'root.{agent_id}',
            log_tags=self._log_tags,
            log_level=log_level,
            log_format=log_format,
            **kwargs
        )

    async def initialize(self) -> None:
        await self._connector.initialize()

        async with asyncio.TaskGroup() as tg:
            tasks: list[asyncio.Task] = list()
            tasks.append(tg.create_task(self._connector.register_cmd_out_channel()))
            tasks.append(tg.create_task(self._connector.subscribe_to_statuses(self._status_callback_generator(self._status_callback))))

    async def start(self):
        await self._connector.start()

    async def stop(self):
        await self._connector.stop()

    def cmd(self,
            cmd: AgentCmd,
            module_types: str | Iterable[str] = ModuleTypes.ALL,
            module_ids: str | Iterable[str] = ModuleTypes.ALL):
        self._connector.cmd(cmd, module_types, module_ids)

    def _status_callback_generator(self, status_callback: Callable[[StatusReport], None]) -> Callable[[StatusReport], None]:
        def callback(status: StatusReport) -> None:
            self.debug(f'Received {status}')
            status_callback(status)

        return callback

    @property
    def _logger_extras(self) -> LoggerExtras | None:
        return LoggerExtras(
            agent_time=self._time.agent,
            mod_time=self._time.module,
            exec_time=str(self._time.exec) if self._time.exec is not None else '-',
            tags=self.log_tag_str
        )



from typing import Any, Iterable, ClassVar

from mhagenta.utils import ModuleTypes, Outbox, ConnType, Message, Observation, State
from mhagenta.core.processes.mha_module import MHAModule, GlobalParams, ModuleBase


class LearnerOutbox(Outbox):
    """Internal communication outbox class for Learner.

    Used to store and process outgoing messages to other modules.

    """
    def request_memories(self, memory_id: str, **kwargs) -> None:
        """Request a collection of memories from a memory structure.

        Args:
            memory_id (str): `module_id` of the relevant memory structure.
            **kwargs: additional keyword arguments to be included in the message.

        """
        self._add(memory_id, ConnType.request, kwargs)

    def send_status(self, ll_reasoner_id: str, learning_status: Any, **kwargs) -> None:
        """Send learning status to a low-level reasoner.

        Args:
            ll_reasoner_id (str): `module_id` of the low-level reasoner to report to.
            learning_status (Any): learning status to report.
            **kwargs: additional keyword arguments to be included in the message.

        """
        body = {'learning_status': learning_status}
        if kwargs:
            body.update(kwargs)
        self._add(ll_reasoner_id, ConnType.send, body, extension='status')

    def send_model(self, ll_reasoner_id: str, model: Any, **kwargs) -> None:
        """Send a learned model to a low-level reasoner.

        Args:
            ll_reasoner_id (str): `module_id` of the relevant low-level reasoner.
            model (Any): model to send.
            **kwargs: additional keyword arguments to be included in the message.

        """
        body = {'model': model}
        if kwargs:
            body.update(kwargs)
        self._add(ll_reasoner_id, ConnType.send, body, extension='model')


LearnerState = State[LearnerOutbox]


class LearnerBase(ModuleBase):
    """Base class for defining Learner behavior (also inherits common methods from `ModuleBase`).

    To implement a custom behavior, override the empty base functions: `on_init`, `step`, `on_first`, `on_last`, and/or
    reactions to messages from other modules.

    """
    module_type: ClassVar[str] = ModuleTypes.LEARNER

    def on_task(self, state: LearnerState, sender: str, task: Any, **kwargs) -> LearnerState:
        """Override to define learner's reaction to receiving a learning task.

        Args:
            state (LearnerState): Learner's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the low-level reasoner that sent the learning task.
            task (Any): received learning task object.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            LearnerState: modified or unaltered internal state of the module.

        """
        return state

    def on_memories(self, state: LearnerState, sender: str, observations: Iterable[Observation], **kwargs) -> LearnerState:
        """Override to define learner's reaction to receiving a collection of memories.

        Args:
            state (LearnerState): Learner's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the memory structure that send the memories.
            observations (Iterable[Observation]): received collection of memories (observation objects).
            **kwargs: additional keyword arguments included in the message.

        Returns:
            LearnerState: modified or unaltered internal state of the module.

        """
        return state

    def on_model_request(self, state: LearnerState, sender: str, **kwargs) -> LearnerState:
        """Override to define learner's reaction to receiving a model request.

        Args:
            state (LearnerState): Learner's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the low-level reasoner that sent the model request.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            LearnerState: modified or unaltered internal state of the module.

        """
        return state


class Learner(MHAModule):
    def __init__(self,
                 global_params: GlobalParams,
                 base: LearnerBase
                 ):
        self._module_id = base.module_id
        self._base = base
        self._directory = global_params.directory

        out_id_channels = list()
        in_id_channels_callbacks = list()

        for ll_reasoner in self._directory.ll_reasoning:
            out_id_channels.append(self.sender_reg_entry(ll_reasoner, ConnType.send, extension='model'))
            out_id_channels.append(self.sender_reg_entry(ll_reasoner, ConnType.send, extension='status'))
            in_id_channels_callbacks.append(self.recipient_reg_entry(ll_reasoner, ConnType.request, self._receive_model_request))
            in_id_channels_callbacks.append(self.recipient_reg_entry(ll_reasoner, ConnType.send, self._receive_task))

        for memory in self._directory.memory:
            out_id_channels.append(self.sender_reg_entry(memory, ConnType.request))
            in_id_channels_callbacks.append(self.recipient_reg_entry(memory, ConnType.send, self._receive_memories))

        super().__init__(
            global_params=global_params,
            base=base,
            out_id_channels=out_id_channels,
            in_id_channel_callbacks=in_id_channels_callbacks,
            outbox_cls=LearnerOutbox
        )

    def _receive_task(self, sender: str, channel: str, msg: Message) -> LearnerState:
        self.info(f'Received a new task {msg.id} from {sender}. Processing...')
        task = msg.body.pop('task')
        update = self._base.on_task(state=self._state, sender=sender, task=task, **msg.body)
        self.debug(f'Finished processing the new task {msg.id}!')
        return update

    def _receive_memories(self, sender: str, channel: str, msg: Message) -> LearnerState:
        self.info(f'Received memories {msg.id} from {sender}. Processing...')
        observations = msg.body.pop('observations')
        update = self._base.on_memories(state=self._state, sender=sender, observations=observations, **msg.body)
        self.debug(f'Finished processing memories {msg.id}!')
        return update

    def _receive_model_request(self, sender: str, channel: str, msg: Message) -> LearnerState:
        self.info(f'Received a model request {msg.id} from {sender}. Processing...')
        update = self._base.on_model_request(state=self._state, sender=sender, **msg.body)
        self.debug(f'Finished processing the model request {msg.id}!')
        return update



from .actuation import Actuator
from .perception import Perceptor
from .low_level import LLReasoner, Learner
from .memory import Memory
from .high_level import Knowledge, HLReasoner, GoalGraph


__all__ = ['Actuator', 'Perceptor', 'LLReasoner', 'Learner', 'Memory', 'Knowledge', 'HLReasoner', 'GoalGraph']



import asyncio
import os
from pathlib import Path
from sys import argv, exit
from typing import *

import dill
from pydantic import BaseModel

from mhagenta.modules import *
from mhagenta.states import *
from mhagenta.utils import ModuleTypes, Observation, ActionStatus
from mhagenta.core.processes import run_agent_module, MHAModule, ModuleBase, GlobalParams
from mhagenta.bases import *


class ModuleSet(BaseModel):
    module_cls: Type[MHAModule]
    base: Type[ModuleBase]


MODULE_NAME_TO_CLASS = {
    ModuleTypes.ACTUATOR: ModuleSet(module_cls=Actuator, base=ActuatorBase),
    ModuleTypes.PERCEPTOR: ModuleSet(module_cls=Perceptor, base=PerceptorBase),
    ModuleTypes.LLREASONER: ModuleSet(module_cls=LLReasoner, base=LLReasonerBase),
    ModuleTypes.LEARNER: ModuleSet(module_cls=Learner, base=LearnerBase),
    ModuleTypes.MEMORY: ModuleSet(module_cls=Memory, base=MemoryBase),
    ModuleTypes.KNOWLEDGE: ModuleSet(module_cls=Knowledge, base=KnowledgeBase),
    ModuleTypes.HLREASONER: ModuleSet(module_cls=HLReasoner, base=HLReasonerBase),
    ModuleTypes.GOALGRAPH: ModuleSet(module_cls=GoalGraph, base=GoalGraphBase)
}


if __name__ == "__main__":
    if len(argv) < 2:
        exit('Expected [params_path] as an argument!')

    file_path = Path(argv[1].replace('\"', ''))
    with open(file_path, 'rb') as f:
        params = dill.load(f)
    os.remove(file_path)
    module_name, params = params['class'], params['kwargs']
    module_data = MODULE_NAME_TO_CLASS[module_name]
    module_cls = module_data.module_cls
    params['base'] = dill.loads(params['base'])
    params['global_params'] = GlobalParams(**params['global_params'])

    exit_reason = asyncio.run(run_agent_module(module_cls, **params))

    print(f'Module {params["base"].module_id} exited, reason: {exit_reason}')
